"use strict";
//
// Copyright 2019-2021 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
//
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CallLogLevel = exports.CallEndedReason = exports.CallState = exports.RingCancelReason = exports.DataMode = exports.HangupType = exports.OpaqueMessage = exports.HangupMessage = exports.BusyMessage = exports.IceCandidateMessage = exports.AnswerMessage = exports.OfferType = exports.OfferMessage = exports.CallingMessage = exports.GroupCall = exports.GroupCallKind = exports.VideoRequest = exports.GroupMemberInfo = exports.RemoteDeviceState = exports.LocalDeviceState = exports.HttpMethod = exports.RingUpdate = exports.CallMessageUrgency = exports.GroupCallEndReason = exports.JoinState = exports.ConnectionState = exports.Call = exports.RingRTCType = exports.ReceivedAudioLevel = exports.NetworkRoute = exports.PeekStatusCodes = exports.callIdFromRingId = exports.callIdFromEra = void 0;
const CallLinks_1 = require("./CallLinks");
const Native_1 = __importDefault(require("./Native"));
exports.callIdFromEra = Native_1.default.callIdFromEra;
function callIdFromRingId(ringId) {
    return {
        low: Number(BigInt.asIntN(32, ringId)),
        high: Number(BigInt.asIntN(32, ringId >> BigInt(32))),
        unsigned: true,
    };
}
exports.callIdFromRingId = callIdFromRingId;
class Config {
}
class NativeCallManager {
    constructor(observer) {
        this.observer = observer;
        this.createCallEndpoint(new Config());
    }
    setConfig(config) {
        this.createCallEndpoint(config);
    }
    createCallEndpoint(config) {
        const fieldTrials = Object.assign({
            'RingRTC-AnyAddressPortsKillSwitch': 'Enabled',
            'WebRTC-Audio-OpusSetSignalVoiceWithDtx': 'Enabled',
        }, config.field_trials);
        /* eslint-disable prefer-template */
        const fieldTrialsString = Object.entries(fieldTrials)
            .map(([k, v]) => `${k}/${v}`)
            .join('/') + '/';
        /* eslint-enable prefer-template */
        Object.defineProperty(this, Native_1.default.callEndpointPropertyKey, {
            configurable: true,
            get() {
                const callEndpoint = Native_1.default.createCallEndpoint(this, fieldTrialsString);
                Object.defineProperty(this, Native_1.default.callEndpointPropertyKey, {
                    configurable: true,
                    value: callEndpoint,
                });
                // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                return callEndpoint;
            },
        });
    }
}
// Mirror methods onto NativeCallManager.
// This is done through direct assignment rather than wrapper methods to avoid indirection.
NativeCallManager.prototype.setSelfUuid = Native_1.default.cm_setSelfUuid;
NativeCallManager.prototype.createOutgoingCall =
    Native_1.default.cm_createOutgoingCall;
NativeCallManager.prototype.proceed = Native_1.default.cm_proceed;
NativeCallManager.prototype.accept = Native_1.default.cm_accept;
NativeCallManager.prototype.ignore = Native_1.default.cm_ignore;
NativeCallManager.prototype.hangup = Native_1.default.cm_hangup;
NativeCallManager.prototype.cancelGroupRing =
    Native_1.default.cm_cancelGroupRing;
NativeCallManager.prototype.signalingMessageSent =
    Native_1.default.cm_signalingMessageSent;
NativeCallManager.prototype.signalingMessageSendFailed =
    Native_1.default.cm_signalingMessageSendFailed;
NativeCallManager.prototype.updateDataMode = Native_1.default.cm_updateDataMode;
NativeCallManager.prototype.receivedOffer = Native_1.default.cm_receivedOffer;
NativeCallManager.prototype.receivedAnswer = Native_1.default.cm_receivedAnswer;
NativeCallManager.prototype.receivedIceCandidates =
    Native_1.default.cm_receivedIceCandidates;
NativeCallManager.prototype.receivedHangup = Native_1.default.cm_receivedHangup;
NativeCallManager.prototype.receivedBusy = Native_1.default.cm_receivedBusy;
NativeCallManager.prototype.receivedCallMessage =
    Native_1.default.cm_receivedCallMessage;
NativeCallManager.prototype.receivedHttpResponse =
    Native_1.default.cm_receivedHttpResponse;
NativeCallManager.prototype.httpRequestFailed =
    Native_1.default.cm_httpRequestFailed;
NativeCallManager.prototype.setOutgoingAudioEnabled =
    Native_1.default.cm_setOutgoingAudioEnabled;
NativeCallManager.prototype.setOutgoingVideoEnabled =
    Native_1.default.cm_setOutgoingVideoEnabled;
NativeCallManager.prototype.setOutgoingVideoIsScreenShare =
    Native_1.default.cm_setOutgoingVideoIsScreenShare;
NativeCallManager.prototype.sendVideoFrame = Native_1.default.cm_sendVideoFrame;
NativeCallManager.prototype.receiveVideoFrame =
    Native_1.default.cm_receiveVideoFrame;
NativeCallManager.prototype.receiveGroupCallVideoFrame =
    Native_1.default.cm_receiveGroupCallVideoFrame;
NativeCallManager.prototype.createGroupCallClient =
    Native_1.default.cm_createGroupCallClient;
NativeCallManager.prototype.createCallLinkCallClient =
    Native_1.default.cm_createCallLinkCallClient;
NativeCallManager.prototype.deleteGroupCallClient =
    Native_1.default.cm_deleteGroupCallClient;
NativeCallManager.prototype.connect = Native_1.default.cm_connect;
NativeCallManager.prototype.join = Native_1.default.cm_join;
NativeCallManager.prototype.leave = Native_1.default.cm_leave;
NativeCallManager.prototype.disconnect = Native_1.default.cm_disconnect;
NativeCallManager.prototype.groupRing = Native_1.default.cm_groupRing;
NativeCallManager.prototype.groupReact = Native_1.default.cm_groupReact;
NativeCallManager.prototype.groupRaiseHand = Native_1.default.cm_groupRaiseHand;
NativeCallManager.prototype.setOutgoingAudioMuted =
    Native_1.default.cm_setOutgoingAudioMuted;
NativeCallManager.prototype.setOutgoingVideoMuted =
    Native_1.default.cm_setOutgoingVideoMuted;
NativeCallManager.prototype.setOutgoingGroupCallVideoIsScreenShare =
    Native_1.default.cm_setOutgoingGroupCallVideoIsScreenShare;
NativeCallManager.prototype.setPresenting = Native_1.default.cm_setPresenting;
NativeCallManager.prototype.resendMediaKeys =
    Native_1.default.cm_resendMediaKeys;
NativeCallManager.prototype.setDataMode = Native_1.default.cm_setDataMode;
NativeCallManager.prototype.requestVideo = Native_1.default.cm_requestVideo;
NativeCallManager.prototype.approveUser = Native_1.default.cm_approveUser;
NativeCallManager.prototype.denyUser = Native_1.default.cm_denyUser;
NativeCallManager.prototype.removeClient = Native_1.default.cm_removeClient;
NativeCallManager.prototype.blockClient = Native_1.default.cm_blockClient;
NativeCallManager.prototype.setGroupMembers =
    Native_1.default.cm_setGroupMembers;
NativeCallManager.prototype.setMembershipProof =
    Native_1.default.cm_setMembershipProof;
NativeCallManager.prototype.readCallLink = Native_1.default.cm_readCallLink;
NativeCallManager.prototype.createCallLink = Native_1.default.cm_createCallLink;
NativeCallManager.prototype.updateCallLink = Native_1.default.cm_updateCallLink;
NativeCallManager.prototype.peekGroupCall = Native_1.default.cm_peekGroupCall;
NativeCallManager.prototype.peekCallLinkCall =
    Native_1.default.cm_peekCallLinkCall;
NativeCallManager.prototype.getAudioInputs = Native_1.default.cm_getAudioInputs;
NativeCallManager.prototype.setAudioInput = Native_1.default.cm_setAudioInput;
NativeCallManager.prototype.getAudioOutputs =
    Native_1.default.cm_getAudioOutputs;
NativeCallManager.prototype.setAudioOutput = Native_1.default.cm_setAudioOutput;
NativeCallManager.prototype.processEvents = Native_1.default.cm_processEvents;
var PeekStatusCodes;
(function (PeekStatusCodes) {
    PeekStatusCodes[PeekStatusCodes["EXPIRED_CALL_LINK"] = 703] = "EXPIRED_CALL_LINK";
    PeekStatusCodes[PeekStatusCodes["INVALID_CALL_LINK"] = 704] = "INVALID_CALL_LINK";
})(PeekStatusCodes || (exports.PeekStatusCodes = PeekStatusCodes = {}));
// In sync with WebRTC's PeerConnection.AdapterType.
// Despite how it looks, this is not an option set.
// A network adapter type can only be one of the listed values.
// And there are a few oddities to note:
// - Cellular means we don't know if it's 2G, 3G, 4G, 5G, ...
//   If we know, it will be one of those corresponding enum values.
//   This means to know if something is cellular or not, you must
//   check all of those values.
// - Default means we don't know the adapter type (like Unknown)
//   but it's because we bound to the default IP address (0.0.0.0)
//   so it's probably the default adapter (wifi if available, for example)
//   This is unlikely to happen in practice.
var NetworkAdapterType;
(function (NetworkAdapterType) {
    NetworkAdapterType[NetworkAdapterType["Unknown"] = 0] = "Unknown";
    NetworkAdapterType[NetworkAdapterType["Ethernet"] = 1] = "Ethernet";
    NetworkAdapterType[NetworkAdapterType["Wifi"] = 2] = "Wifi";
    NetworkAdapterType[NetworkAdapterType["Cellular"] = 4] = "Cellular";
    NetworkAdapterType[NetworkAdapterType["Vpn"] = 8] = "Vpn";
    NetworkAdapterType[NetworkAdapterType["Loopback"] = 16] = "Loopback";
    NetworkAdapterType[NetworkAdapterType["Default"] = 32] = "Default";
    NetworkAdapterType[NetworkAdapterType["Cellular2G"] = 64] = "Cellular2G";
    NetworkAdapterType[NetworkAdapterType["Cellular3G"] = 128] = "Cellular3G";
    NetworkAdapterType[NetworkAdapterType["Cellular4G"] = 256] = "Cellular4G";
    NetworkAdapterType[NetworkAdapterType["Cellular5G"] = 512] = "Cellular5G";
})(NetworkAdapterType || (NetworkAdapterType = {}));
// Information about the network route being used for sending audio/video/data
class NetworkRoute {
    constructor() {
        this.localAdapterType = NetworkAdapterType.Unknown;
    }
}
exports.NetworkRoute = NetworkRoute;
class ReceivedAudioLevel {
    constructor(demuxId, level) {
        this.demuxId = demuxId;
        this.level = level;
    }
}
exports.ReceivedAudioLevel = ReceivedAudioLevel;
function normalizeAudioLevel(raw) {
    return raw / 32767;
}
class Requests {
    constructor() {
        this._resolveById = new Map();
        this._nextId = 1;
    }
    add() {
        const id = this._nextId++;
        const promise = new Promise((resolve, _reject) => {
            this._resolveById.set(id, resolve);
        });
        return [id, promise];
    }
    resolve(id, response) {
        const resolve = this._resolveById.get(id);
        if (!resolve) {
            return false;
        }
        resolve(response);
        this._resolveById.delete(id);
        return true;
    }
}
class CallInfo {
    constructor(isVideoCall, receivedAtCounter) {
        this.isVideoCall = isVideoCall;
        this.receivedAtCounter = receivedAtCounter;
    }
}
class RingRTCType {
    getCallInfoKey(callId) {
        // CallId is u64 so use a string key instead.
        // Note that the representation is not padded, so we include a separator.
        // Otherwise {1, 123} and {11, 23} would have the same key.
        // (We could use Long.toString as well, but it doesn't matter what the key is.)
        return `${callId.high} ${callId.low}`;
    }
    constructor() {
        // Set by UX
        this.handleOutgoingSignaling = null;
        this.handleIncomingCall = null;
        this.handleStartCall = null;
        this.handleAutoEndedIncomingCallRequest = null;
        this.handleLogMessage = null;
        this.handleSendHttpRequest = null;
        this.handleSendCallMessage = null;
        this.handleSendCallMessageToGroup = null;
        this.handleGroupCallRingUpdate = null;
        this.callManager = new NativeCallManager(this);
        this._call = null;
        this._groupCallByClientId = new Map();
        this._peekRequests = new Requests();
        this._callLinkRequests = new Requests();
        this._callInfoByCallId = new Map();
    }
    setConfig(config) {
        this.callManager.setConfig(config);
    }
    // Called by UX
    setSelfUuid(uuid) {
        this.callManager.setSelfUuid(uuid);
    }
    // Called by UX
    startOutgoingCall(remoteUserId, isVideoCall, localDeviceId) {
        const callId = this.callManager.createOutgoingCall(remoteUserId, isVideoCall, localDeviceId);
        const isIncoming = false;
        const call = new Call(this.callManager, remoteUserId, callId, isIncoming, isVideoCall, CallState.Prering);
        this._call = call;
        // We won't actually send anything until the remote side accepts.
        call.outgoingAudioEnabled = true;
        call.outgoingVideoEnabled = isVideoCall;
        return call;
    }
    // Called by UX
    cancelGroupRing(groupId, ringId, reason) {
        sillyDeadlockProtection(() => {
            this.callManager.cancelGroupRing(groupId, ringId.toString(), reason);
        });
    }
    // Called by Rust
    onStartOutgoingCall(remoteUserId, callId) {
        const call = this._call;
        if (!call || call.remoteUserId !== remoteUserId) {
            return;
        }
        call.callId = callId;
        const handleStartCall = this.handleStartCall;
        if (!handleStartCall) {
            call.ignore();
            return;
        }
        handleStartCall(call)
            .then(result => {
            if (!result) {
                this.logWarn('RingRTC.handleStartCall failed for outgoing call. Call ignored.');
                call.ignore();
            }
        })
            .catch(e => {
            this.logError(`RingRTC.handleStartCall exception: ${e}`);
            call.ignore();
        });
    }
    // Called by Rust
    onStartIncomingCall(remoteUserId, callId, isVideoCall) {
        // Temporary: Force hangup in all glare scenarios until handled gracefully.
        // In case of a glare loser, an incoming call will be generated right
        // after the outgoing call is ended. In that case, ignore it once.
        if (this._call &&
            (this._call.endedReason === CallEndedReason.Glare ||
                this._call.endedReason === CallEndedReason.ReCall)) {
            this._call.endedReason = undefined;
            // EVIL HACK: We are the "loser" of a glare collision and have ended the outgoing call
            // and are now receiving the incoming call from the remote side (the "winner").
            // However, the Desktop client has a bug where it re-orders the events so that
            // instead of seeing ("outgoing call ended", "incoming call"), it sees
            // ("incoming call", "call ended") and it gets messed up.
            // The solution?  Delay processing the incoming call.
            setTimeout(() => {
                this.onStartIncomingCall(remoteUserId, callId, isVideoCall);
            }, 500);
            return;
        }
        const isIncoming = true;
        const call = new Call(this.callManager, remoteUserId, callId, isIncoming, isVideoCall, CallState.Prering);
        const { handleIncomingCall, handleStartCall } = this;
        if (!handleIncomingCall || !handleStartCall) {
            call.ignore();
            return;
        }
        this._call = call;
        handleIncomingCall(call)
            .then(success => {
            if (!success) {
                this.logWarn('RingRTC.handleIncomingCall failed for incoming call. Call ignored.');
                call.ignore();
            }
            else {
                handleStartCall(call)
                    .then(innerSuccess => {
                    if (!innerSuccess) {
                        this.logWarn('RingRTC.handleStartCall failed for incoming call. Call ignored.');
                        call.ignore();
                    }
                })
                    .catch(e => {
                    this.logError(`RingRTC.handleStartCall exception: ${e}`);
                    call.ignore();
                });
            }
        })
            .catch(e => {
            this.logError(`RingRTC.handleIncomingCall exception: ${e}`);
            call.ignore();
        });
    }
    proceed(callId, settings) {
        sillyDeadlockProtection(() => {
            this.callManager.proceed(callId, settings.iceServer.username || '', settings.iceServer.password || '', settings.iceServer.hostname || '', settings.iceServer.urls, settings.hideIp, settings.dataMode, settings.audioLevelsIntervalMillis || 0);
        });
    }
    // Called by Rust
    onCallState(remoteUserId, state) {
        const call = this._call;
        if (!call || call.remoteUserId !== remoteUserId) {
            return;
        }
        call.state = state;
    }
    // Called by Rust
    onCallEnded(remoteUserId, callId, reason, ageSec) {
        const callInfo = this._callInfoByCallId.get(this.getCallInfoKey(callId));
        const { isVideoCall, receivedAtCounter } = callInfo || {
            isVideoCall: false,
            receivedAtCounter: undefined,
        };
        this._callInfoByCallId.delete(this.getCallInfoKey(callId));
        const call = this._call;
        if (call && reason == CallEndedReason.ReceivedOfferWithGlare) {
            // The current call is the outgoing call.
            // The ended call is the incoming call.
            // We're the "winner", so ignore the incoming call and keep going with the outgoing call.
            return;
        }
        if (call &&
            (reason === CallEndedReason.Glare || reason === CallEndedReason.ReCall)) {
            // The current call is the outgoing call.
            // The ended call is the outgoing call.
            // We're the "loser", so end the outgoing/current call and wait for a new incoming call.
            // (proceeded down to the code below)
        }
        // If there is no call or the remoteUserId doesn't match that of the current
        // call, or if one of the "receive offer while already in a call or because
        // it expired" reasons are provided, don't end the current call, because
        // there isn't one for this Ended notification, just update the call history.
        // If the incoming call ends while in the prering state, also immediately
        // update the call history because it is just a replay of messages.
        if (!call ||
            call.remoteUserId !== remoteUserId ||
            reason === CallEndedReason.ReceivedOfferWhileActive ||
            reason === CallEndedReason.ReceivedOfferExpired ||
            (call.state === CallState.Prering && call.isIncoming)) {
            if (this.handleAutoEndedIncomingCallRequest) {
                this.handleAutoEndedIncomingCallRequest(callId, remoteUserId, reason, ageSec, isVideoCall, receivedAtCounter);
            }
            if (call && call.state === CallState.Prering && call.isIncoming) {
                // Set the state to Ended without triggering a state update since we
                // already notified the client.
                call.endedReason = reason;
                call.setCallEnded();
            }
            return;
        }
        // Send the end reason first because setting the state triggers
        // call.handleStateChanged, which may look at call.endedReason.
        call.endedReason = reason;
        call.state = CallState.Ended;
    }
    onRemoteVideoEnabled(remoteUserId, enabled) {
        const call = this._call;
        if (!call || call.remoteUserId !== remoteUserId) {
            return;
        }
        call.remoteVideoEnabled = enabled;
        if (call.handleRemoteVideoEnabled) {
            call.handleRemoteVideoEnabled();
        }
    }
    onRemoteSharingScreen(remoteUserId, enabled) {
        const call = this._call;
        if (!call || call.remoteUserId !== remoteUserId) {
            return;
        }
        call.remoteSharingScreen = enabled;
        if (call.handleRemoteSharingScreen) {
            call.handleRemoteSharingScreen();
        }
    }
    onNetworkRouteChanged(remoteUserId, localNetworkAdapterType) {
        const call = this._call;
        if (!call || call.remoteUserId !== remoteUserId) {
            return;
        }
        call.networkRoute.localAdapterType = localNetworkAdapterType;
        if (call.handleNetworkRouteChanged) {
            call.handleNetworkRouteChanged();
        }
    }
    onAudioLevels(remoteUserId, capturedLevel, receivedLevel) {
        const call = this._call;
        if (!call || call.remoteUserId !== remoteUserId) {
            return;
        }
        call.outgoingAudioLevel = normalizeAudioLevel(capturedLevel);
        call.remoteAudioLevel = normalizeAudioLevel(receivedLevel);
        if (call.handleAudioLevels) {
            call.handleAudioLevels();
        }
    }
    onLowBandwidthForVideo(remoteUserId, recovered) {
        const call = this._call;
        if (!call || call.remoteUserId !== remoteUserId) {
            return;
        }
        if (call.handleLowBandwidthForVideo) {
            call.handleLowBandwidthForVideo(recovered);
        }
    }
    renderVideoFrame(width, height, buffer) {
        const call = this._call;
        if (!call) {
            return;
        }
        if (this._call?.renderVideoFrame) {
            this._call?.renderVideoFrame(width, height, buffer);
        }
    }
    // Called by Rust
    onSendOffer(remoteUserId, remoteDeviceId, callId, broadcast, offerType, opaque) {
        const message = new CallingMessage();
        message.offer = new OfferMessage(callId, offerType, opaque);
        this.sendSignaling(remoteUserId, remoteDeviceId, callId, broadcast, message);
    }
    // Called by Rust
    onSendAnswer(remoteUserId, remoteDeviceId, callId, broadcast, opaque) {
        const message = new CallingMessage();
        message.answer = new AnswerMessage(callId, opaque);
        this.sendSignaling(remoteUserId, remoteDeviceId, callId, broadcast, message);
    }
    // Called by Rust
    onSendIceCandidates(remoteUserId, remoteDeviceId, callId, broadcast, candidates) {
        const message = new CallingMessage();
        message.iceCandidates = [];
        for (const candidate of candidates) {
            const copy = new IceCandidateMessage(callId, candidate);
            message.iceCandidates.push(copy);
        }
        this.sendSignaling(remoteUserId, remoteDeviceId, callId, broadcast, message);
    }
    // Called by Rust
    onSendHangup(remoteUserId, remoteDeviceId, callId, broadcast, hangupType, deviceId) {
        const message = new CallingMessage();
        message.hangup = new HangupMessage(callId, hangupType, deviceId || 0);
        this.sendSignaling(remoteUserId, remoteDeviceId, callId, broadcast, message);
    }
    // Called by Rust
    onSendBusy(remoteUserId, remoteDeviceId, callId, broadcast) {
        const message = new CallingMessage();
        message.busy = new BusyMessage(callId);
        this.sendSignaling(remoteUserId, remoteDeviceId, callId, broadcast, message);
    }
    sendSignaling(remoteUserId, remoteDeviceId, callId, broadcast, message) {
        if (!broadcast) {
            message.destinationDeviceId = remoteDeviceId;
        }
        (async () => {
            if (this.handleOutgoingSignaling) {
                const signalingResult = await this.handleOutgoingSignaling(remoteUserId, message);
                if (signalingResult) {
                    this.callManager.signalingMessageSent(callId);
                }
                else {
                    this.callManager.signalingMessageSendFailed(callId);
                }
            }
            else {
                this.callManager.signalingMessageSendFailed(callId);
            }
        })().catch(e => this.logError(e.toString()));
    }
    // Call Links
    /**
     * Asynchronous request to get information about a call link.
     *
     * @param sfuUrl - the URL to use when accessing the SFU
     * @param authCredentialPresentation - a serialized CallLinkAuthCredentialPresentation
     * @param linkRootKey - the root key for the call link
     *
     * Expected failure codes include:
     * - 404: the room does not exist (or expired so long ago that it has been removed from the server)
     */
    readCallLink(sfuUrl, authCredentialPresentation, linkRootKey) {
        const [requestId, promise] = this._callLinkRequests.add();
        // Response comes back via handleCallLinkResponse
        sillyDeadlockProtection(() => {
            this.callManager.readCallLink(requestId, sfuUrl, authCredentialPresentation, linkRootKey.bytes);
        });
        return promise;
    }
    /**
     * Asynchronous request to create a new call link.
     *
     * This request is idempotent; if it fails due to a network issue, it is safe to retry.
     *
     * @example
     * const linkKey = CallLinkRootKey.generate();
     * const adminPasskey = CallLinkRootKey.generateAdminPasskey();
     * const roomId = linkKey.deriveRoomId();
     * const credential = requestCreateCredentialFromChatServer(roomId); // using libsignal
     * const secretParams = CallLinkSecretParams.deriveFromRootKey(linkKey.bytes);
     * const credentialPresentation = credential.present(roomId, secretParams).serialize();
     * const serializedPublicParams = secretParams.getPublicParams().serialize();
     * const result = await RingRTC.createCallLink(sfuUrl, credentialPresentation, linkKey, adminPasskey, serializedPublicParams);
     * if (result.success) {
     *   const state = result.value;
     *   // In actuality you may not want to do this until the user clicks Done.
     *   saveToDatabase(linkKey.bytes, adminPasskey, state);
     *   syncToOtherDevices(linkKey.bytes, adminPasskey);
     * } else {
     *   switch (result.errorStatusCode) {
     *   case 409:
     *     // The room already exists (and isn't yours), i.e. you've hit a 1-in-a-billion conflict.
     *     // Fall through to kicking the user out to try again later.
     *   default:
     *     // Unexpected error, kick the user out for now.
     *   }
     * }
     *
     * @param sfuUrl - the URL to use when accessing the SFU
     * @param createCredentialPresentation - a serialized CreateCallLinkCredentialPresentation
     * @param linkRootKey - the root key for the call link
     * @param adminPasskey - the arbitrary passkey to use for the new room
     * @param callLinkPublicParams - the serialized CallLinkPublicParams for the new room
     */
    createCallLink(sfuUrl, createCredentialPresentation, linkRootKey, adminPasskey, callLinkPublicParams) {
        const [requestId, promise] = this._callLinkRequests.add();
        // Response comes back via handleCallLinkResponse
        sillyDeadlockProtection(() => {
            this.callManager.createCallLink(requestId, sfuUrl, createCredentialPresentation, linkRootKey.bytes, adminPasskey, callLinkPublicParams);
        });
        return promise;
    }
    /**
     * Asynchronous request to update a call link's name.
     *
     * Possible failure codes include:
     * - 401: the room does not exist (and this is the wrong API to create a new room)
     * - 403: the admin passkey is incorrect
     *
     * This request is idempotent; if it fails due to a network issue, it is safe to retry.
     *
     * @param sfuUrl - the URL to use when accessing the SFU
     * @param authCredentialPresentation - a serialized CallLinkAuthCredentialPresentation
     * @param linkRootKey - the root key for the call link
     * @param adminPasskey - the passkey specified when the link was created
     * @param newName - the new name to use
     */
    updateCallLinkName(sfuUrl, authCredentialPresentation, linkRootKey, adminPasskey, newName) {
        const [requestId, promise] = this._callLinkRequests.add();
        // Response comes back via handleCallLinkResponse
        sillyDeadlockProtection(() => {
            this.callManager.updateCallLink(requestId, sfuUrl, authCredentialPresentation, linkRootKey.bytes, adminPasskey, newName, undefined, undefined);
        });
        return promise;
    }
    /**
     * Asynchronous request to update a call link's restrictions.
     *
     * Possible failure codes include:
     * - 401: the room does not exist (and this is the wrong API to create a new room)
     * - 403: the admin passkey is incorrect
     * - 409: the room is currently in use, so restrictions cannot be changed at the moment
     *
     * This request is idempotent; if it fails due to a network issue, it is safe to retry.
     *
     * @param sfuUrl - the URL to use when accessing the SFU
     * @param authCredentialPresentation - a serialized CallLinkAuthCredentialPresentation
     * @param linkRootKey - the root key for the call link
     * @param adminPasskey - the passkey specified when the link was created
     * @param restrictions - the new restrictions to use
     */
    updateCallLinkRestrictions(sfuUrl, authCredentialPresentation, linkRootKey, adminPasskey, restrictions) {
        const [requestId, promise] = this._callLinkRequests.add();
        // Response comes back via handleCallLinkResponse
        sillyDeadlockProtection(() => {
            this.callManager.updateCallLink(requestId, sfuUrl, authCredentialPresentation, linkRootKey.bytes, adminPasskey, undefined, restrictions, undefined);
        });
        return promise;
    }
    /**
     * Asynchronous request to revoke or un-revoke a call link.
     *
     * Possible failure codes include:
     * - 401: the room does not exist (and this is the wrong API to create a new room)
     * - 403: the admin passkey is incorrect
     *
     * This request is idempotent; if it fails due to a network issue, it is safe to retry.
     *
     * @param sfuUrl - the URL to use when accessing the SFU
     * @param authCredentialPresentation - a serialized CallLinkAuthCredentialPresentation
     * @param linkRootKey - the root key for the call link
     * @param adminPasskey - the passkey specified when the link was created
     * @param revoked - whether the link should now be revoked
     */
    updateCallLinkRevocation(sfuUrl, authCredentialPresentation, linkRootKey, adminPasskey, revoked) {
        const [requestId, promise] = this._callLinkRequests.add();
        // Response comes back via handleCallLinkResponse
        sillyDeadlockProtection(() => {
            this.callManager.updateCallLink(requestId, sfuUrl, authCredentialPresentation, linkRootKey.bytes, adminPasskey, undefined, undefined, revoked);
        });
        return promise;
    }
    // HTTP callbacks
    receivedHttpResponse(requestId, status, body) {
        sillyDeadlockProtection(() => {
            try {
                this.callManager.receivedHttpResponse(requestId, status, body);
            }
            catch {
                // We may not have an active connection any more.
                // In which case it doesn't matter
            }
        });
    }
    httpRequestFailed(requestId, debugInfo) {
        sillyDeadlockProtection(() => {
            try {
                this.callManager.httpRequestFailed(requestId, debugInfo);
            }
            catch {
                // We may not have an active connection any more.
                // In which case it doesn't matter
            }
        });
    }
    // Group Calls
    // Called by UX
    getGroupCall(groupId, sfuUrl, hkdfExtraInfo, audioLevelsIntervalMillis, observer) {
        const clientId = this.callManager.createGroupCallClient(groupId, sfuUrl, hkdfExtraInfo, audioLevelsIntervalMillis || 0);
        const groupCall = new GroupCall(GroupCallKind.SignalGroup, this.callManager, observer, clientId);
        this._groupCallByClientId.set(groupCall.clientId, groupCall);
        return groupCall;
    }
    // Called by UX
    getCallLinkCall(sfuUrl, authCredentialPresentation, rootKey, adminPasskey, hkdfExtraInfo, audioLevelsIntervalMillis, observer) {
        const clientId = this.callManager.createCallLinkCallClient(sfuUrl, authCredentialPresentation, rootKey.bytes, adminPasskey, hkdfExtraInfo, audioLevelsIntervalMillis || 0);
        const groupCall = new GroupCall(GroupCallKind.CallLink, this.callManager, observer, clientId);
        this._groupCallByClientId.set(groupCall.clientId, groupCall);
        return groupCall;
    }
    // Called by UX
    // Returns a list of user IDs
    peekGroupCall(sfuUrl, membershipProof, groupMembers) {
        const [requestId, promise] = this._peekRequests.add();
        // Response comes back via handlePeekResponse
        sillyDeadlockProtection(() => {
            this.callManager.peekGroupCall(requestId, sfuUrl, membershipProof, groupMembers);
        });
        return promise.then(result => {
            if (result.success) {
                return result.value;
            }
            else {
                return {
                    devices: [],
                    deviceCount: 0,
                    deviceCountIncludingPendingDevices: 0,
                    deviceCountExcludingPendingDevices: 0,
                    pendingUsers: [],
                };
            }
        });
    }
    // Called by UX
    peekCallLinkCall(sfuUrl, authCredentialPresentation, rootKey) {
        const [requestId, promise] = this._peekRequests.add();
        // Response comes back via handlePeekResponse
        sillyDeadlockProtection(() => {
            this.callManager.peekCallLinkCall(requestId, sfuUrl, authCredentialPresentation, rootKey.bytes);
        });
        return promise;
    }
    // Called by Rust
    requestMembershipProof(clientId) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('requestMembershipProof(): GroupCall not found in map!');
                return;
            }
            groupCall.requestMembershipProof();
        });
    }
    // Called by Rust
    requestGroupMembers(clientId) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('requestGroupMembers(): GroupCall not found in map!');
                return;
            }
            groupCall.requestGroupMembers();
        });
    }
    // Called by Rust
    handleConnectionStateChanged(clientId, connectionState) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('handleConnectionStateChanged(): GroupCall not found in map!');
                return;
            }
            groupCall.handleConnectionStateChanged(connectionState);
        });
    }
    // Called by Rust
    handleJoinStateChanged(clientId, joinState, demuxId) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('handleJoinStateChanged(): GroupCall not found in map!');
                return;
            }
            groupCall.handleJoinStateChanged(joinState, demuxId);
        });
    }
    // Called by Rust
    handleNetworkRouteChanged(clientId, localNetworkAdapterType) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('handleNetworkRouteChanged(): GroupCall not found in map!');
                return;
            }
            groupCall.handleNetworkRouteChanged(localNetworkAdapterType);
        });
    }
    // Called by Rust
    handleAudioLevels(clientId, capturedLevel, receivedLevels) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (groupCall) {
                groupCall.handleAudioLevels(capturedLevel, receivedLevels);
            }
        });
    }
    // Called by Rust
    handleLowBandwidthForVideo(clientId, recovered) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (groupCall) {
                groupCall.handleLowBandwidthForVideo(recovered);
            }
        });
    }
    // Called by Rust
    handleReactions(clientId, reactions) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (groupCall) {
                groupCall.handleReactions(reactions);
            }
        });
    }
    // Called by Rust
    handleRaisedHands(clientId, raisedHands) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (groupCall) {
                groupCall.handleRaisedHands(raisedHands);
            }
        });
    }
    // Called by Rust
    handleRemoteDevicesChanged(clientId, remoteDeviceStates) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('handleRemoteDevicesChanged(): GroupCall not found in map!');
                return;
            }
            groupCall.handleRemoteDevicesChanged(remoteDeviceStates);
        });
    }
    // Called by Rust
    handlePeekChanged(clientId, info) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('handlePeekChanged(): GroupCall not found in map!');
                return;
            }
            groupCall.handlePeekChanged(info);
        });
    }
    // Called by Rust
    handlePeekResponse(requestId, statusCode, info) {
        sillyDeadlockProtection(() => {
            let result;
            if (info) {
                result = { success: true, value: info };
            }
            else {
                result = { success: false, errorStatusCode: statusCode };
            }
            if (!this._peekRequests.resolve(requestId, result)) {
                this.logWarn(`Invalid request ID for handlePeekResponse: ${requestId}`);
            }
        });
    }
    // Called by Rust
    handleCallLinkResponse(requestId, statusCode, state) {
        sillyDeadlockProtection(() => {
            // Recreate the state so that we have the correct prototype, in case we add more methods to CallLinkState.
            let result;
            if (state) {
                let restrictions;
                switch (state.rawRestrictions) {
                    case 0:
                        restrictions = CallLinks_1.CallLinkRestrictions.None;
                        break;
                    case 1:
                        restrictions = CallLinks_1.CallLinkRestrictions.AdminApproval;
                        break;
                    default:
                        restrictions = CallLinks_1.CallLinkRestrictions.Unknown;
                        break;
                }
                result = {
                    success: true,
                    value: new CallLinks_1.CallLinkState(state.name, restrictions, state.revoked, state.expiration),
                };
            }
            else {
                result = { success: false, errorStatusCode: statusCode };
            }
            if (!this._callLinkRequests.resolve(requestId, result)) {
                this.logWarn(`Invalid request ID for handleCallLinkResponse: ${requestId}`);
            }
        });
    }
    // Called by Rust
    handleEnded(clientId, reason) {
        sillyDeadlockProtection(() => {
            const groupCall = this._groupCallByClientId.get(clientId);
            if (!groupCall) {
                this.logError('handleEnded(): GroupCall not found in map!');
                return;
            }
            this._groupCallByClientId.delete(clientId);
            groupCall.handleEnded(reason);
        });
    }
    // Called by Rust
    groupCallRingUpdate(groupId, ringIdString, sender, state) {
        sillyDeadlockProtection(() => {
            if (this.handleGroupCallRingUpdate) {
                const ringId = BigInt(ringIdString);
                this.handleGroupCallRingUpdate(groupId, ringId, sender, state);
            }
            else {
                this.logError('RingRTC.handleGroupCallRingUpdate is not set!');
            }
        });
    }
    // Called by Rust
    onLogMessage(level, fileName, line, message) {
        if (this.handleLogMessage) {
            this.handleLogMessage(level, fileName, line, message);
        }
    }
    // Called from here
    logError(message) {
        this.onLogMessage(CallLogLevel.Error, 'Service.ts', 0, message);
    }
    // Called from here
    logWarn(message) {
        this.onLogMessage(CallLogLevel.Warn, 'Service.ts', 0, message);
    }
    // Called from here
    logInfo(message) {
        this.onLogMessage(CallLogLevel.Info, 'Service.ts', 0, message);
    }
    // Called by MessageReceiver
    handleCallingMessage(remoteUserId, remoteUuid, remoteDeviceId, localDeviceId, messageAgeSec, messageReceivedAtCounter, message, senderIdentityKey, receiverIdentityKey) {
        if (message.destinationDeviceId &&
            message.destinationDeviceId !== localDeviceId) {
            // Drop the message as it isn't for this device, handleIgnoredCall() is not needed.
            return;
        }
        if (message.offer?.callId) {
            const callId = message.offer.callId;
            const opaque = toBuffer(message.offer.opaque);
            // opaque is required. sdp is obsolete, but it might still come with opaque.
            if (!opaque) {
                // TODO: Remove once the proto is updated to only support opaque and require it.
                this.logError('handleCallingMessage(): opaque not received for offer, remote should update');
                return;
            }
            const offerType = message.offer.type || OfferType.AudioCall;
            // Save the call details for later when the call is ended.
            const callInfo = new CallInfo(offerType === OfferType.VideoCall, messageReceivedAtCounter);
            this._callInfoByCallId.set(this.getCallInfoKey(callId), callInfo);
            this.callManager.receivedOffer(remoteUserId, remoteDeviceId, localDeviceId, messageAgeSec, callId, offerType, opaque, senderIdentityKey, receiverIdentityKey);
        }
        if (message.answer?.callId) {
            const callId = message.answer.callId;
            const opaque = toBuffer(message.answer.opaque);
            // opaque is required. sdp is obsolete, but it might still come with opaque.
            if (!opaque) {
                // TODO: Remove once the proto is updated to only support opaque and require it.
                this.logError('handleCallingMessage(): opaque not received for answer, remote should update');
                return;
            }
            this.callManager.receivedAnswer(remoteUserId, remoteDeviceId, callId, opaque, senderIdentityKey, receiverIdentityKey);
        }
        if (message.iceCandidates && message.iceCandidates.length > 0) {
            // We assume they all have the same .callId
            const callId = message.iceCandidates[0].callId;
            // We have to copy them to do the .toArrayBuffer() thing.
            const candidates = [];
            for (const candidate of message.iceCandidates) {
                const copy = toBuffer(candidate.opaque);
                if (copy) {
                    candidates.push(copy);
                }
                else {
                    // TODO: Remove once the proto is updated to only support opaque and require it.
                    this.logError('handleCallingMessage(): opaque not received for ice candidate, remote should update');
                    continue;
                }
            }
            if (candidates.length == 0) {
                this.logWarn('handleCallingMessage(): No ice candidates in ice message, remote should update');
                return;
            }
            if (!callId) {
                this.logWarn('handleCallingMessage(): No call ID in ice message');
                return;
            }
            this.callManager.receivedIceCandidates(remoteUserId, remoteDeviceId, callId, candidates);
        }
        if (message.hangup?.callId) {
            const callId = message.hangup.callId;
            const hangupType = message.hangup.type || HangupType.Normal;
            const hangupDeviceId = message.hangup.deviceId || null;
            this.callManager.receivedHangup(remoteUserId, remoteDeviceId, callId, hangupType, hangupDeviceId);
        }
        if (message.busy?.callId) {
            const callId = message.busy.callId;
            this.callManager.receivedBusy(remoteUserId, remoteDeviceId, callId);
        }
        if (message.opaque) {
            if (remoteUuid == null) {
                this.logError('handleCallingMessage(): opaque message received without UUID!');
                return;
            }
            const data = toBuffer(message.opaque.data);
            if (data == undefined) {
                this.logError('handleCallingMessage(): opaque message received without data!');
                return;
            }
            this.callManager.receivedCallMessage(remoteUuid, remoteDeviceId, localDeviceId, data, messageAgeSec);
        }
    }
    // Called by Rust
    sendHttpRequest(requestId, url, method, headers, body) {
        if (this.handleSendHttpRequest) {
            this.handleSendHttpRequest(requestId, url, method, headers, body);
        }
        else {
            this.logError('RingRTC.handleSendHttpRequest is not set!');
        }
    }
    // Called by Rust
    sendCallMessage(recipientUuid, message, urgency) {
        if (this.handleSendCallMessage) {
            this.handleSendCallMessage(recipientUuid, message, urgency);
        }
        else {
            this.logError('RingRTC.handleSendCallMessage is not set!');
        }
    }
    // Called by Rust
    sendCallMessageToGroup(groupId, message, urgency) {
        if (this.handleSendCallMessageToGroup) {
            this.handleSendCallMessageToGroup(groupId, message, urgency);
        }
        else {
            this.logError('RingRTC.handleSendCallMessageToGroup is not set!');
        }
    }
    // These are convenience methods.  One could use the Call class instead.
    get call() {
        return this._call;
    }
    getCall(callId) {
        const call = this.call;
        if (call &&
            call.callId.high === callId.high &&
            call.callId.low === callId.low) {
            return call;
        }
        return null;
    }
    accept(callId, asVideoCall) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.accept();
        call.outgoingAudioEnabled = true;
        call.outgoingVideoEnabled = asVideoCall;
    }
    decline(callId) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.decline();
    }
    ignore(callId) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.ignore();
    }
    hangup(callId) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.hangup();
    }
    setOutgoingAudio(callId, enabled) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.outgoingAudioEnabled = enabled;
    }
    setOutgoingVideo(callId, enabled) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.outgoingVideoEnabled = enabled;
    }
    setOutgoingVideoIsScreenShare(callId, isScreenShare) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.outgoingVideoIsScreenShare = isScreenShare;
    }
    setVideoCapturer(callId, capturer) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.videoCapturer = capturer;
    }
    setVideoRenderer(callId, renderer) {
        const call = this.getCall(callId);
        if (!call) {
            return;
        }
        call.videoRenderer = renderer;
    }
    getAudioInputs() {
        return this.callManager.getAudioInputs();
    }
    setAudioInput(index) {
        this.callManager.setAudioInput(index);
    }
    getAudioOutputs() {
        return this.callManager.getAudioOutputs();
    }
    setAudioOutput(index) {
        this.callManager.setAudioOutput(index);
    }
}
exports.RingRTCType = RingRTCType;
class Call {
    constructor(callManager, remoteUserId, callId, isIncoming, isVideoCall, state) {
        this._outgoingAudioEnabled = false;
        this._outgoingVideoEnabled = false;
        this._outgoingVideoIsScreenShare = false;
        this._remoteVideoEnabled = false;
        this.outgoingAudioLevel = 0;
        this.remoteAudioLevel = 0;
        this.remoteSharingScreen = false;
        this.networkRoute = new NetworkRoute();
        this._videoCapturer = null;
        this._videoRenderer = null;
        this._callManager = callManager;
        this._remoteUserId = remoteUserId;
        this.callId = callId;
        this._isIncoming = isIncoming;
        this._isVideoCall = isVideoCall;
        this._state = state;
    }
    get remoteUserId() {
        return this._remoteUserId;
    }
    get isIncoming() {
        return this._isIncoming;
    }
    get isVideoCall() {
        return this._isVideoCall;
    }
    get state() {
        return this._state;
    }
    set state(state) {
        if (state == this._state) {
            return;
        }
        this._state = state;
        this.enableOrDisableCapturer();
        this.enableOrDisableRenderer();
        if (this.handleStateChanged) {
            this.handleStateChanged();
        }
    }
    setCallEnded() {
        this._state = CallState.Ended;
    }
    set videoCapturer(capturer) {
        this._videoCapturer = capturer;
        this.enableOrDisableCapturer();
    }
    set videoRenderer(renderer) {
        this._videoRenderer = renderer;
        this.enableOrDisableRenderer();
    }
    accept() {
        this._callManager.accept(this.callId);
    }
    decline() {
        this.hangup();
    }
    ignore() {
        this._callManager.ignore(this.callId);
    }
    hangup() {
        // This is a little faster than waiting for the
        // change in call state to come back.
        if (this._videoCapturer) {
            this._videoCapturer.disable();
        }
        if (this._videoRenderer) {
            this._videoRenderer.disable();
        }
        // This assumes we only have one active call.
        sillyDeadlockProtection(() => {
            this._callManager.hangup();
        });
    }
    get outgoingAudioEnabled() {
        return this._outgoingAudioEnabled;
    }
    set outgoingAudioEnabled(enabled) {
        this._outgoingAudioEnabled = enabled;
        // This assumes we only have one active call.
        sillyDeadlockProtection(() => {
            this._callManager.setOutgoingAudioEnabled(enabled);
        });
    }
    get outgoingVideoEnabled() {
        return this._outgoingVideoEnabled;
    }
    set outgoingVideoEnabled(enabled) {
        this._outgoingVideoEnabled = enabled;
        this.enableOrDisableCapturer();
    }
    set outgoingVideoIsScreenShare(isScreenShare) {
        // This assumes we only have one active call.
        this._outgoingVideoIsScreenShare = isScreenShare;
        sillyDeadlockProtection(() => {
            this._callManager.setOutgoingVideoIsScreenShare(isScreenShare);
        });
    }
    get remoteVideoEnabled() {
        return this._remoteVideoEnabled;
    }
    set remoteVideoEnabled(enabled) {
        this._remoteVideoEnabled = enabled;
        this.enableOrDisableRenderer();
    }
    // With this method, a Call is a VideoFrameSender
    sendVideoFrame(width, height, format, buffer) {
        // This assumes we only have one active call.
        this._callManager.sendVideoFrame(width, height, format, buffer);
    }
    // With this method, a Call is a VideoFrameSource
    receiveVideoFrame(buffer, maxWidth, maxHeight) {
        // This assumes we only have one active call.
        return this._callManager.receiveVideoFrame(buffer, maxWidth, maxHeight);
    }
    enableOrDisableCapturer() {
        if (!this._videoCapturer) {
            return;
        }
        if (!this.outgoingVideoEnabled) {
            this._videoCapturer.disable();
            if (this.state === CallState.Accepted) {
                this.setOutgoingVideoEnabled(false);
            }
            return;
        }
        switch (this.state) {
            case CallState.Prering:
            case CallState.Ringing:
                this._videoCapturer.enableCapture();
                break;
            case CallState.Accepted:
                this._videoCapturer.enableCaptureAndSend(this);
                this.setOutgoingVideoEnabled(true);
                if (this._outgoingVideoIsScreenShare) {
                    // Make sure the status gets sent.
                    this.outgoingVideoIsScreenShare = true;
                }
                break;
            case CallState.Reconnecting:
                this._videoCapturer.enableCaptureAndSend(this);
                // Don't send status until we're reconnected.
                break;
            case CallState.Ended:
                this._videoCapturer.disable();
                break;
            default:
        }
    }
    setOutgoingVideoEnabled(enabled) {
        sillyDeadlockProtection(() => {
            try {
                this._callManager.setOutgoingVideoEnabled(enabled);
            }
            catch {
                // We may not have an active connection any more.
                // In which case it doesn't matter
            }
        });
    }
    updateDataMode(dataMode) {
        sillyDeadlockProtection(() => {
            try {
                this._callManager.updateDataMode(dataMode);
            }
            catch {
                // We may not have an active connection any more.
                // In which case it doesn't matter
            }
        });
    }
    enableOrDisableRenderer() {
        if (!this._videoRenderer) {
            return;
        }
        if (!this.remoteVideoEnabled) {
            this._videoRenderer.disable();
            return;
        }
        switch (this.state) {
            case CallState.Prering:
            case CallState.Ringing:
                this._videoRenderer.disable();
                break;
            case CallState.Accepted:
            case CallState.Reconnecting:
                this._videoRenderer.enable(this);
                break;
            case CallState.Ended:
                this._videoRenderer.disable();
                break;
            default:
        }
    }
}
exports.Call = Call;
// Represents the connection state to a media server for a group call.
var ConnectionState;
(function (ConnectionState) {
    ConnectionState[ConnectionState["NotConnected"] = 0] = "NotConnected";
    ConnectionState[ConnectionState["Connecting"] = 1] = "Connecting";
    ConnectionState[ConnectionState["Connected"] = 2] = "Connected";
    ConnectionState[ConnectionState["Reconnecting"] = 3] = "Reconnecting";
})(ConnectionState || (exports.ConnectionState = ConnectionState = {}));
// Represents whether or not a user is joined to a group call and can exchange media.
var JoinState;
(function (JoinState) {
    JoinState[JoinState["NotJoined"] = 0] = "NotJoined";
    JoinState[JoinState["Joining"] = 1] = "Joining";
    JoinState[JoinState["Pending"] = 2] = "Pending";
    JoinState[JoinState["Joined"] = 3] = "Joined";
})(JoinState || (exports.JoinState = JoinState = {}));
// If not ended purposely by the user, gives the reason why a group call ended.
var GroupCallEndReason;
(function (GroupCallEndReason) {
    // Normal events
    GroupCallEndReason[GroupCallEndReason["DeviceExplicitlyDisconnected"] = 0] = "DeviceExplicitlyDisconnected";
    GroupCallEndReason[GroupCallEndReason["ServerExplicitlyDisconnected"] = 1] = "ServerExplicitlyDisconnected";
    GroupCallEndReason[GroupCallEndReason["DeniedRequestToJoinCall"] = 2] = "DeniedRequestToJoinCall";
    GroupCallEndReason[GroupCallEndReason["RemovedFromCall"] = 3] = "RemovedFromCall";
    // Things that can go wrong
    GroupCallEndReason[GroupCallEndReason["CallManagerIsBusy"] = 4] = "CallManagerIsBusy";
    GroupCallEndReason[GroupCallEndReason["SfuClientFailedToJoin"] = 5] = "SfuClientFailedToJoin";
    GroupCallEndReason[GroupCallEndReason["FailedToCreatePeerConnectionFactory"] = 6] = "FailedToCreatePeerConnectionFactory";
    GroupCallEndReason[GroupCallEndReason["FailedToNegotiateSrtpKeys"] = 7] = "FailedToNegotiateSrtpKeys";
    GroupCallEndReason[GroupCallEndReason["FailedToCreatePeerConnection"] = 8] = "FailedToCreatePeerConnection";
    GroupCallEndReason[GroupCallEndReason["FailedToStartPeerConnection"] = 9] = "FailedToStartPeerConnection";
    GroupCallEndReason[GroupCallEndReason["FailedToUpdatePeerConnection"] = 10] = "FailedToUpdatePeerConnection";
    GroupCallEndReason[GroupCallEndReason["FailedToSetMaxSendBitrate"] = 11] = "FailedToSetMaxSendBitrate";
    GroupCallEndReason[GroupCallEndReason["IceFailedWhileConnecting"] = 12] = "IceFailedWhileConnecting";
    GroupCallEndReason[GroupCallEndReason["IceFailedAfterConnected"] = 13] = "IceFailedAfterConnected";
    GroupCallEndReason[GroupCallEndReason["ServerChangedDemuxId"] = 14] = "ServerChangedDemuxId";
    GroupCallEndReason[GroupCallEndReason["HasMaxDevices"] = 15] = "HasMaxDevices";
})(GroupCallEndReason || (exports.GroupCallEndReason = GroupCallEndReason = {}));
var CallMessageUrgency;
(function (CallMessageUrgency) {
    CallMessageUrgency[CallMessageUrgency["Droppable"] = 0] = "Droppable";
    CallMessageUrgency[CallMessageUrgency["HandleImmediately"] = 1] = "HandleImmediately";
})(CallMessageUrgency || (exports.CallMessageUrgency = CallMessageUrgency = {}));
var RingUpdate;
(function (RingUpdate) {
    /// The sender is trying to ring this user.
    RingUpdate[RingUpdate["Requested"] = 0] = "Requested";
    /// The sender tried to ring this user, but it's been too long.
    RingUpdate[RingUpdate["ExpiredRequest"] = 1] = "ExpiredRequest";
    /// Call was accepted elsewhere by a different device.
    RingUpdate[RingUpdate["AcceptedOnAnotherDevice"] = 2] = "AcceptedOnAnotherDevice";
    /// Call was declined elsewhere by a different device.
    RingUpdate[RingUpdate["DeclinedOnAnotherDevice"] = 3] = "DeclinedOnAnotherDevice";
    /// This device is currently on a different call.
    RingUpdate[RingUpdate["BusyLocally"] = 4] = "BusyLocally";
    /// A different device is currently on a different call.
    RingUpdate[RingUpdate["BusyOnAnotherDevice"] = 5] = "BusyOnAnotherDevice";
    /// The sender cancelled the ring request.
    RingUpdate[RingUpdate["CancelledByRinger"] = 6] = "CancelledByRinger";
})(RingUpdate || (exports.RingUpdate = RingUpdate = {}));
// HTTP request methods.
var HttpMethod;
(function (HttpMethod) {
    HttpMethod[HttpMethod["Get"] = 0] = "Get";
    HttpMethod[HttpMethod["Put"] = 1] = "Put";
    HttpMethod[HttpMethod["Post"] = 2] = "Post";
    HttpMethod[HttpMethod["Delete"] = 3] = "Delete";
})(HttpMethod || (exports.HttpMethod = HttpMethod = {}));
// The local device state for a group call.
class LocalDeviceState {
    constructor() {
        this.connectionState = ConnectionState.NotConnected;
        this.joinState = JoinState.NotJoined;
        // By default audio and video are muted.
        this.audioMuted = true;
        this.videoMuted = true;
        this.audioLevel = 0;
        this.presenting = false;
        this.sharingScreen = false;
        this.networkRoute = new NetworkRoute();
    }
}
exports.LocalDeviceState = LocalDeviceState;
// All remote devices in a group call and their associated state.
class RemoteDeviceState {
    constructor(demuxId, userId, mediaKeysReceived) {
        this.demuxId = demuxId;
        this.userId = userId;
        this.mediaKeysReceived = mediaKeysReceived;
        this.audioLevel = 0;
        this.isHigherResolutionPending = false;
    }
}
exports.RemoteDeviceState = RemoteDeviceState;
// Used to communicate the group membership to RingRTC for a group call.
class GroupMemberInfo {
    constructor(userId, userIdCipherText) {
        this.userId = userId;
        this.userIdCipherText = userIdCipherText;
    }
}
exports.GroupMemberInfo = GroupMemberInfo;
// Used for the application to communicate the actual resolutions of
// each device in a group call to RingRTC and the SFU.
class VideoRequest {
    constructor(demuxId, width, height, framerate) {
        this.demuxId = demuxId;
        this.width = width;
        this.height = height;
        this.framerate = framerate;
    }
}
exports.VideoRequest = VideoRequest;
var GroupCallKind;
(function (GroupCallKind) {
    GroupCallKind[GroupCallKind["SignalGroup"] = 0] = "SignalGroup";
    GroupCallKind[GroupCallKind["CallLink"] = 1] = "CallLink";
})(GroupCallKind || (exports.GroupCallKind = GroupCallKind = {}));
class GroupCall {
    get clientId() {
        return this._clientId;
    }
    // Called by UI via RingRTC object
    constructor(kind, callManager, observer, clientId) {
        this._kind = kind;
        this._callManager = callManager;
        this._observer = observer;
        this._clientId = clientId;
        this._localDeviceState = new LocalDeviceState();
    }
    getKind() {
        return this._kind;
    }
    // Called by UI
    connect() {
        this._callManager.connect(this._clientId);
    }
    // Called by UI
    join() {
        this._callManager.join(this._clientId);
    }
    // Called by UI
    leave() {
        this._callManager.leave(this._clientId);
    }
    // Called by UI
    disconnect() {
        this._callManager.disconnect(this._clientId);
    }
    // Called by UI
    getLocalDeviceState() {
        return this._localDeviceState;
    }
    // Called by UI
    getRemoteDeviceStates() {
        return this._remoteDeviceStates;
    }
    // Called by UI
    getPeekInfo() {
        return this._peekInfo;
    }
    // Called by UI
    setOutgoingAudioMuted(muted) {
        this._localDeviceState.audioMuted = muted;
        this._callManager.setOutgoingAudioMuted(this._clientId, muted);
        this._observer.onLocalDeviceStateChanged(this);
    }
    // Called by UI
    react(value) {
        this._callManager.groupReact(this._clientId, value);
    }
    // Called by UI
    raiseHand(raise) {
        this._callManager.groupRaiseHand(this._clientId, raise);
    }
    // Called by UI
    setOutgoingVideoMuted(muted) {
        this._localDeviceState.videoMuted = muted;
        this._callManager.setOutgoingVideoMuted(this._clientId, muted);
        this._observer.onLocalDeviceStateChanged(this);
    }
    // Called by UI
    setPresenting(presenting) {
        this._localDeviceState.presenting = presenting;
        this._callManager.setPresenting(this._clientId, presenting);
        this._observer.onLocalDeviceStateChanged(this);
    }
    // Called by UI
    setOutgoingVideoIsScreenShare(isScreenShare) {
        this._localDeviceState.sharingScreen = isScreenShare;
        this._callManager.setOutgoingGroupCallVideoIsScreenShare(this._clientId, isScreenShare);
        this._observer.onLocalDeviceStateChanged(this);
    }
    // Called by UI
    ringAll() {
        this._callManager.groupRing(this._clientId, undefined);
    }
    // Called by UI
    resendMediaKeys() {
        this._callManager.resendMediaKeys(this._clientId);
    }
    // Called by UI
    setDataMode(dataMode) {
        this._callManager.setDataMode(this._clientId, dataMode);
    }
    // Called by UI
    requestVideo(resolutions, activeSpeakerHeight) {
        this._callManager.requestVideo(this._clientId, resolutions, activeSpeakerHeight);
    }
    // Called by UI
    approveUser(otherUserId) {
        this._callManager.approveUser(this._clientId, otherUserId);
    }
    // Called by UI
    denyUser(otherUserId) {
        this._callManager.denyUser(this._clientId, otherUserId);
    }
    // Called by UI
    removeClient(otherClientDemuxId) {
        this._callManager.removeClient(this._clientId, otherClientDemuxId);
    }
    // Called by UI
    blockClient(otherClientDemuxId) {
        this._callManager.blockClient(this._clientId, otherClientDemuxId);
    }
    // Called by UI
    setGroupMembers(members) {
        this._callManager.setGroupMembers(this._clientId, members);
    }
    // Called by UI
    setMembershipProof(proof) {
        this._callManager.setMembershipProof(this._clientId, proof);
    }
    // Called by Rust via RingRTC object
    requestMembershipProof() {
        this._observer.requestMembershipProof(this);
    }
    // Called by Rust via RingRTC object
    requestGroupMembers() {
        this._observer.requestGroupMembers(this);
    }
    // Called by Rust via RingRTC object
    handleConnectionStateChanged(connectionState) {
        this._localDeviceState.connectionState = connectionState;
        this._observer.onLocalDeviceStateChanged(this);
    }
    // Called by Rust via RingRTC object
    handleJoinStateChanged(joinState, demuxId) {
        this._localDeviceState.joinState = joinState;
        // Don't set to undefined after we leave so we can still know the demuxId after we leave.
        if (demuxId != undefined) {
            this._localDeviceState.demuxId = demuxId;
        }
        this._observer.onLocalDeviceStateChanged(this);
    }
    // Called by Rust via RingRTC object
    handleNetworkRouteChanged(localNetworkAdapterType) {
        this._localDeviceState.networkRoute.localAdapterType =
            localNetworkAdapterType;
        this._observer.onLocalDeviceStateChanged(this);
    }
    handleAudioLevels(capturedLevel, receivedLevels) {
        this._localDeviceState.audioLevel = normalizeAudioLevel(capturedLevel);
        if (this._remoteDeviceStates != undefined) {
            for (const received of receivedLevels) {
                for (const remoteDeviceState of this._remoteDeviceStates) {
                    if (remoteDeviceState.demuxId == received.demuxId) {
                        remoteDeviceState.audioLevel = normalizeAudioLevel(received.level);
                    }
                }
            }
        }
        this._observer.onAudioLevels(this);
    }
    handleLowBandwidthForVideo(recovered) {
        this._observer.onLowBandwidthForVideo(this, recovered);
    }
    handleReactions(reactions) {
        this._observer.onReactions(this, reactions);
    }
    handleRaisedHands(raisedHands) {
        this._observer.onRaisedHands(this, raisedHands);
    }
    // Called by Rust via RingRTC object
    handleRemoteDevicesChanged(remoteDeviceStates) {
        // We don't get aspect ratios from RingRTC, so make sure to copy them over.
        for (const noo of remoteDeviceStates) {
            const old = this._remoteDeviceStates?.find(old => old.demuxId == noo.demuxId);
            noo.videoAspectRatio = old?.videoAspectRatio;
        }
        this._remoteDeviceStates = remoteDeviceStates;
        this._observer.onRemoteDeviceStatesChanged(this);
    }
    // Called by Rust via RingRTC object
    handlePeekChanged(info) {
        this._peekInfo = info;
        this._observer.onPeekChanged(this);
    }
    // Called by Rust via RingRTC object
    handleEnded(reason) {
        this._observer.onEnded(this, reason);
        this._callManager.deleteGroupCallClient(this._clientId);
    }
    // With this, a GroupCall is a VideoFrameSender
    sendVideoFrame(width, height, format, buffer) {
        // This assumes we only have one active call.
        this._callManager.sendVideoFrame(width, height, format, buffer);
    }
    // With this, a GroupCall can provide a VideoFrameSource for each remote device.
    getVideoSource(remoteDemuxId) {
        return new GroupCallVideoFrameSource(this._callManager, this, remoteDemuxId);
    }
    // Called by the GroupCallVideoFrameSource when it receives a video frame.
    setRemoteAspectRatio(remoteDemuxId, aspectRatio) {
        const remoteDevice = this._remoteDeviceStates?.find(device => device.demuxId == remoteDemuxId);
        if (!!remoteDevice && remoteDevice.videoAspectRatio != aspectRatio) {
            remoteDevice.videoAspectRatio = aspectRatio;
            this._observer.onRemoteDeviceStatesChanged(this);
        }
    }
}
exports.GroupCall = GroupCall;
// Implements VideoSource for use in CanvasVideoRenderer
class GroupCallVideoFrameSource {
    constructor(callManager, groupCall, remoteDemuxId // Uint32
    ) {
        this._callManager = callManager;
        this._groupCall = groupCall;
        this._remoteDemuxId = remoteDemuxId;
    }
    receiveVideoFrame(buffer, maxWidth, maxHeight) {
        // This assumes we only have one active call.
        const frame = this._callManager.receiveGroupCallVideoFrame(this._groupCall.clientId, this._remoteDemuxId, buffer, maxWidth, maxHeight);
        if (frame) {
            const [width, height] = frame;
            this._groupCall.setRemoteAspectRatio(this._remoteDemuxId, width / height);
        }
        return frame;
    }
}
function toBuffer(pbab) {
    if (!pbab) {
        return pbab;
    }
    if (pbab instanceof Buffer) {
        return pbab;
    }
    return Buffer.from(pbab.toArrayBuffer());
}
class CallingMessage {
}
exports.CallingMessage = CallingMessage;
class OfferMessage {
    constructor(callId, type, opaque) {
        this.callId = callId;
        this.type = type;
        this.opaque = opaque;
    }
}
exports.OfferMessage = OfferMessage;
var OfferType;
(function (OfferType) {
    OfferType[OfferType["AudioCall"] = 0] = "AudioCall";
    OfferType[OfferType["VideoCall"] = 1] = "VideoCall";
})(OfferType || (exports.OfferType = OfferType = {}));
class AnswerMessage {
    constructor(callId, opaque) {
        this.callId = callId;
        this.opaque = opaque;
    }
}
exports.AnswerMessage = AnswerMessage;
class IceCandidateMessage {
    constructor(callId, opaque) {
        this.callId = callId;
        this.opaque = opaque;
    }
}
exports.IceCandidateMessage = IceCandidateMessage;
class BusyMessage {
    constructor(callId) {
        this.callId = callId;
    }
}
exports.BusyMessage = BusyMessage;
class HangupMessage {
    constructor(callId, type, deviceId) {
        this.callId = callId;
        this.type = type;
        this.deviceId = deviceId;
    }
}
exports.HangupMessage = HangupMessage;
class OpaqueMessage {
}
exports.OpaqueMessage = OpaqueMessage;
var HangupType;
(function (HangupType) {
    HangupType[HangupType["Normal"] = 0] = "Normal";
    HangupType[HangupType["Accepted"] = 1] = "Accepted";
    HangupType[HangupType["Declined"] = 2] = "Declined";
    HangupType[HangupType["Busy"] = 3] = "Busy";
    HangupType[HangupType["NeedPermission"] = 4] = "NeedPermission";
})(HangupType || (exports.HangupType = HangupType = {}));
var DataMode;
(function (DataMode) {
    DataMode[DataMode["Low"] = 0] = "Low";
    DataMode[DataMode["Normal"] = 1] = "Normal";
})(DataMode || (exports.DataMode = DataMode = {}));
/// Describes why a ring was cancelled.
var RingCancelReason;
(function (RingCancelReason) {
    /// The user explicitly clicked "Decline".
    RingCancelReason[RingCancelReason["DeclinedByUser"] = 0] = "DeclinedByUser";
    /// The device is busy with another call.
    RingCancelReason[RingCancelReason["Busy"] = 1] = "Busy";
})(RingCancelReason || (exports.RingCancelReason = RingCancelReason = {}));
var CallState;
(function (CallState) {
    CallState["Prering"] = "idle";
    CallState["Ringing"] = "ringing";
    CallState["Accepted"] = "connected";
    CallState["Reconnecting"] = "connecting";
    CallState["Ended"] = "ended";
})(CallState || (exports.CallState = CallState = {}));
var CallEndedReason;
(function (CallEndedReason) {
    CallEndedReason["LocalHangup"] = "LocalHangup";
    CallEndedReason["RemoteHangup"] = "RemoteHangup";
    CallEndedReason["RemoteHangupNeedPermission"] = "RemoteHangupNeedPermission";
    CallEndedReason["Declined"] = "Declined";
    CallEndedReason["Busy"] = "Busy";
    CallEndedReason["Glare"] = "Glare";
    CallEndedReason["ReCall"] = "ReCall";
    CallEndedReason["ReceivedOfferExpired"] = "ReceivedOfferExpired";
    CallEndedReason["ReceivedOfferWhileActive"] = "ReceivedOfferWhileActive";
    CallEndedReason["ReceivedOfferWithGlare"] = "ReceivedOfferWithGlare";
    CallEndedReason["SignalingFailure"] = "SignalingFailure";
    CallEndedReason["GlareFailure"] = "GlareFailure";
    CallEndedReason["ConnectionFailure"] = "ConnectionFailure";
    CallEndedReason["InternalFailure"] = "InternalFailure";
    CallEndedReason["Timeout"] = "Timeout";
    CallEndedReason["AcceptedOnAnotherDevice"] = "AcceptedOnAnotherDevice";
    CallEndedReason["DeclinedOnAnotherDevice"] = "DeclinedOnAnotherDevice";
    CallEndedReason["BusyOnAnotherDevice"] = "BusyOnAnotherDevice";
})(CallEndedReason || (exports.CallEndedReason = CallEndedReason = {}));
var CallLogLevel;
(function (CallLogLevel) {
    CallLogLevel[CallLogLevel["Off"] = 0] = "Off";
    CallLogLevel[CallLogLevel["Error"] = 1] = "Error";
    CallLogLevel[CallLogLevel["Warn"] = 2] = "Warn";
    CallLogLevel[CallLogLevel["Info"] = 3] = "Info";
    CallLogLevel[CallLogLevel["Debug"] = 4] = "Debug";
    CallLogLevel[CallLogLevel["Trace"] = 5] = "Trace";
})(CallLogLevel || (exports.CallLogLevel = CallLogLevel = {}));
function sillyDeadlockProtection(f) {
    void (async () => {
        // This is a silly way of preventing a deadlock.
        // eslint-disable-next-line @typescript-eslint/await-thenable
        await 0;
        f();
    })();
}
//# sourceMappingURL=Service.js.map